from django.db import models
from django.urls import reverse
from rest_framework import serializers
from rest_framework.request import Request

from djangocms_rest.serializers.placeholder import PlaceholderRelationFieldSerializer


class PageTreeSerializer(serializers.ListSerializer):
    def __init__(self, tree, *args, **kwargs):
        self.tree = tree
        super().__init__(tree.get(None, []), *args, **kwargs)

    def tree_to_representation(self, item):
        repr = self.child.to_representation(item)
        if item.page.node in self.tree:
            repr["children"] = [
                self.tree_to_representation(child) for child in self.tree[item.page.node]
            ]
        return repr

    def to_representation(self, data):
        """
        List of object instances -> List of dicts of primitive datatypes.
        """
        # Dealing with nested relationships, data can be a Manager,
        # so, first get a queryset from the Manager if needed
        iterable = data.all() if isinstance(data, models.manager.BaseManager) else data

        return [
            self.tree_to_representation(item) for item in iterable
        ]


class PageContentSerializer(serializers.Serializer):
    title = serializers.CharField(max_length=255)
    page_title = serializers.CharField(max_length=255)
    menu_title = serializers.CharField(max_length=255)
    meta_description = serializers.CharField()
    redirect = serializers.CharField(max_length=2048)
    absolute_url = serializers.URLField(max_length=200)

    placeholders = serializers.JSONField()
    path = serializers.CharField(max_length=200)

    is_home = serializers.BooleanField()
    in_navigation = serializers.BooleanField()
    soft_root = serializers.BooleanField()
    template = serializers.CharField(max_length=100)
    xframe_options = serializers.CharField(max_length=50)
    limit_visibility_in_menu = serializers.BooleanField()

    language = serializers.CharField(max_length=10)
    languages = serializers.ListSerializer(
        child=serializers.CharField(), allow_empty=True, required=False
    )

    def __init__(self, request: Request, *args, **kwargs) -> None:
        self.request = request
        super().__init__(*args, **kwargs)

    @classmethod
    def many_init(cls, request, instances, *args, **kwargs):
        kwargs['child'] = cls(request)
        if kwargs.pop("as_tree", True):
            tree = {}
            for instance in instances:
                if instance.page.node.parent in tree:
                    tree[instance.page.node.parent].append(instance)
                else:
                    tree[instance.page.node.parent] = [instance]
            return PageTreeSerializer(tree, *args, **kwargs)
        return serializers.ListSerializer(instances, *args, **kwargs)

    def to_representation(self, page_content):
        declared_slots = [
            placeholder.slot
            for placeholder in page_content.page.get_declared_placeholders()
        ]
        placeholders = [
            placeholder
            for placeholder in page_content.page.get_placeholders(page_content.language)
            if placeholder.slot in declared_slots
        ]

        return {
            "title": page_content.title,
            "page_title": page_content.page_title or page_content.title,
            "menu_title": page_content.menu_title or page_content.title,
            "meta_description": page_content.meta_description,
            "redirect": page_content.redirect,
            "placeholders": PlaceholderRelationFieldSerializer(
                self.request,
                page_content,
                placeholders,
                page_content.language,
            ).data,
            "in_navigation": page_content.in_navigation,
            "soft_root": page_content.soft_root,
            "template": page_content.template,
            "xframe_options": page_content.xframe_options,
            "limit_visibility_in_menu": page_content.limit_visibility_in_menu,
            "language": page_content.language,
            "absolute_url": page_content.page.get_absolute_url(
                page_content.language
            ),
            "path": f"{self.request.scheme}://{self.request.get_host()}" + reverse(
                "cms-page-root",
                args=(page_content.language,)
            ) if page_content.page.is_home else f"{self.request.scheme}://{self.request.get_host()}" + reverse(
                "cms-page-detail",
                args=(page_content.language, page_content.page.get_path(page_content.language),)
            ),
            "is_home": page_content.page.is_home,
            "languages": page_content.page.languages.split(","),
        }



import time

from cms.cache.placeholder import (
    _get_placeholder_cache_key,
    _get_placeholder_cache_version_key,
)
from cms.models import Placeholder
from cms.plugin_rendering import BaseRenderer, ContentRenderer
from cms.utils.conf import get_cms_setting
from cms.utils.plugins import get_plugins
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.shortcuts import get_current_site
from django.db import models
from django.template.defaulttags import now
from django.urls import reverse
from rest_framework import serializers
from rest_framework.request import Request
from sekizai.context import SekizaiContext
from sekizai.helpers import get_varname


def _get_placeholder_cache_version(placeholder, lang, site_id):
    """
    Gets the (placeholder x lang)'s current version and vary-on header-names
    list, if present, otherwise resets to («timestamp», []).
    """
    from django.core.cache import cache

    key = _get_placeholder_cache_version_key(placeholder, lang, site_id) + ":rest"
    cached = cache.get(key)
    if cached:
        version, vary_on_list = cached
    else:
        version = int(time.time() * 1000000)
        vary_on_list = []
        _set_placeholder_cache_version(
            placeholder, lang, site_id, version, vary_on_list
        )
    return version, vary_on_list


def _set_placeholder_cache_version(
    placeholder, lang, site_id, version, vary_on_list=None, duration=None
):
    """
    Sets the (placeholder x lang)'s version and vary-on header-names list.
    """
    from django.core.cache import cache

    key = _get_placeholder_cache_version_key(placeholder, lang, site_id) + ":rest"

    if not version or version < 1:
        version = int(time.time() * 1000000)

    if vary_on_list is None:
        vary_on_list = []

    cache.set(key, (version, vary_on_list), duration)


def set_placeholder_rest_cache(placeholder, lang, site_id, content, request):
    """
    Sets the (correct) placeholder cache with the rendered placeholder.
    """
    from django.core.cache import cache

    key = _get_placeholder_cache_key(placeholder, lang, site_id, request) + ":rest"

    duration = min(
        get_cms_setting("CACHE_DURATIONS")["content"],
        placeholder.get_cache_expiration(request, now()),
    )
    cache.set(key, content, duration)
    # "touch" the cache-version, so that it stays as fresh as this content.
    version, vary_on_list = _get_placeholder_cache_version(placeholder, lang, site_id)
    _set_placeholder_cache_version(
        placeholder, lang, site_id, version, vary_on_list, duration=duration
    )


def get_placeholder_rest_cache(placeholder, lang, site_id, request):
    """
    Returns the placeholder from cache respecting the placeholder's
    VARY headers.
    """
    from django.core.cache import cache

    key = (
        _get_placeholder_cache_key(placeholder, lang, site_id, request, soft=True)
        + ":rest"
    )
    content = cache.get(key)
    return content


class PlaceholderRenderer(BaseRenderer):
    """
    The `PlaceholderRenderer` class is a custom renderer that renders a placeholder object.
    """

    def placeholder_cache_is_enabled(self):
        if not get_cms_setting("PLACEHOLDER_CACHE"):
            return False
        if self.request.user.is_staff:
            return False
        return True

    def render_placeholder(self, placeholder, context, language, use_cache=False):
        context.update({"request": self.request})
        if use_cache and placeholder.cache_placeholder:
            use_cache = self.placeholder_cache_is_enabled()
        else:
            use_cache = False

        if use_cache:
            cached_value = get_placeholder_rest_cache(
                placeholder,
                lang=language,
                site_id=get_current_site(self.request).pk,
                request=self.request,
            )
        else:
            cached_value = None

        if cached_value is not None:
            # User has opted to use the cache
            # and there is something in the cache
            return cached_value["content"]

        plugin_content = self.render_plugins(
            placeholder,
            language=language,
            context=context,
        )

        if use_cache:
            set_placeholder_rest_cache(
                placeholder,
                lang=language,
                site_id=get_current_site(self.request).pk,
                content=plugin_content,
                request=self.request,
            )

        if placeholder.pk not in self._rendered_placeholders:
            # First time this placeholder is rendered
            self._rendered_placeholders[placeholder.pk] = plugin_content

        return plugin_content

    def render_plugins(
        self, placeholder: Placeholder, language: str, context: dict
    ) -> list:
        plugins = get_plugins(
            self.request,
            placeholder=placeholder,
            lang=language,
            template=None,
        )

        def render_children(plugins):
            for plugin in plugins:
                plugin_content = self.render_plugin(plugin, context)
                if getattr(plugin, "child_plugin_instances", None):
                    plugin_content["children"] = render_children(
                        plugin.child_plugin_instances
                    )
                if plugin_content:
                    yield plugin_content

        return list(render_children(plugins))

    def render_plugin(self, instance, context):
        instance, plugin = instance.get_plugin_instance()
        if not instance:
            return None
        if hasattr(instance, "serializer") and issubclass(
            instance.serializer, serializers.Serializer
        ):
            json = instance.serializer(instance, context=context).data
        if hasattr(instance, "serialize"):
            json = instance.serialize(context=context)

        class PluginSerializer(serializers.ModelSerializer):
            class Meta:
                model = instance.__class__
                exclude = (
                    "id",
                    "placeholder",
                    "language",
                    "position",
                    "creation_date",
                    "changed_date",
                    "parent",
                )

        json = PluginSerializer(instance, context=context).data
        return json


class PlaceholderSerializer(serializers.Serializer):
    slot = serializers.CharField()
    label = serializers.CharField()
    language = serializers.CharField()
    content = serializers.ListSerializer(
        child=serializers.JSONField(), allow_empty=True, required=False
    )

    def __init__(
        self, request: Request, placeholder: Placeholder, language: str, *args, **kwargs
    ):
        renderer = PlaceholderRenderer(request)
        placeholder.content = renderer.render_placeholder(
            placeholder,
            context={},
            language=language,
            use_cache=True,
        )
        if request.GET.get("html", False):
            html = self.render_html(request, placeholder, language)
            for key, value in html.items():
                if not hasattr(placeholder, key):
                    setattr(placeholder, key, value)
                    self.fields[key] = serializers.CharField()
        placeholder.label = placeholder.get_label()
        placeholder.language = language
        super().__init__(placeholder, *args, **kwargs)

    def render_html(self, request, placeholder, language):
        content_renderer = ContentRenderer(request)
        context = SekizaiContext({"request": request, "LANGUAGE_CODE": language})
        content = content_renderer.render_placeholder(
            placeholder,
            context=context,
            language=language,
            use_cache=True,
        )
        sekizai_blocks = context[get_varname()]

        return {
            "html": content,
            **{key: "".join(value) for key, value in sekizai_blocks.items() if value},
        }


class PlaceholderRelationFieldSerializer(serializers.Serializer):
    def __init__(self, request: Request, instance: models.Model, placeholders, language: str, *args, **kwargs) -> None:
        self.placeholders = placeholders
        self.language: str = language
        super().__init__(instance,*args, **kwargs)
        self.host: str = f"{request.scheme}://{request.get_host()}"

        for placeholder in self.placeholders:
            self.fields[placeholder.slot] = serializers.JSONField()

    def to_representation(self, instance):
        content_type_id = ContentType.objects.get_for_model(
            instance.__class__
        ).pk

        return (
            {
                placeholder.slot: self.host
                + reverse(
                    "cms-placeholder-detail",
                    args=(
                        self.language,
                        content_type_id,
                        instance.pk,
                        placeholder.slot,
                    ),
                )
                for placeholder in self.placeholders
            }
            if instance
            else {}
        )



from cms.models import Page, PageUrl, Placeholder, PageContent
from cms.utils.conf import get_languages
from cms.utils.i18n import get_language_tuple
from cms.utils.page_permissions import user_can_view_page
from django.contrib.sites.shortcuts import get_current_site
from django.http import Http404
from django.urls import reverse
from rest_framework.response import Response
from rest_framework.views import APIView as DRFAPIView

from djangocms_rest.serializers.pageserializer import PageContentSerializer
from djangocms_rest.serializers.placeholder import PlaceholderSerializer


class APIView(DRFAPIView):
    # This is a base class for all API views. It sets the allowed methods to GET and OPTIONS.
    http_method_names = ("get", "options")


class LanguageList(APIView):
    """
    List of languages available for the site. For each language the API returns the
    link to the list of pages for that languages.
    """

    def get(self, request, format=None):
        languages = get_languages().get(get_current_site(request).id, None)
        if languages is None:
            raise Http404
        for conf in languages:
            conf["pages"] = f"{request.scheme}://{request.get_host()}" + reverse(
                "cms-page-list", args=(conf["code"],)
            )
        return Response(languages)


class PageList(APIView):
    """List of all pages on this site for a given language."""

    def get(self, request, language, format=None):
        site = get_current_site(request)
        allowed_languages = [lang[0] for lang in get_language_tuple(site.id)]
        if language not in allowed_languages:
            raise Http404
        qs = Page.objects.filter(node__site=site)
        if request.user.is_anonymous:
            qs = qs.filter(login_required=False)
        pages = (
            page.get_content_obj(language, fallback=True)
            for page in qs
            if user_can_view_page(request.user, page) and page.get_content_obj(language, fallback=True)
        )
        serializer = PageContentSerializer(request, pages, many=True, read_only=True)
        return Response(serializer.data)


class PageDetail(APIView):
    """
    Retrieve a page instance. The page instance includes the placeholders and
    their links to retrieve dynamic content.
    """

    def get_object(self, site, path):
        page_urls = (
            PageUrl.objects.get_for_site(site)
            .filter(path=path)
            .select_related("page__node")
        )
        page_urls = list(page_urls)  # force queryset evaluation to save 1 query
        try:
            page = page_urls[0].page
        except IndexError:
            raise Http404
        else:
            page.urls_cache = {url.language: url for url in page_urls}
        return page

    def get(self, request, language, path="", format=None):
        site = get_current_site(request)
        allowed_languages = [lang[0] for lang in get_language_tuple(site.pk)]
        if language not in allowed_languages:
            raise Http404
        page = self.get_object(site, path)

        # Check if the user has permission to view the page
        if not user_can_view_page(request.user, page):
            raise Http404

        serializer = PageContentSerializer(
            request, page.get_content_obj(language, fallback=True), read_only=True
        )
        return Response(serializer.data)


class PlaceholderDetail(APIView):
    """Placeholder contain the dynamic content. This view retrieves the content as a
    structured nested object.

    Attributes:

    - "slot": The slot name of the placeholder.
    - "content": The content of the placeholder as a nested JSON tree
    - "language": The language of the content
    - "label": The verbose label of the placeholder

    Optional (if the get parameter `?html=1` is added to the API url):
    - "html": The content rendered as html. Sekizai blocks such as "js" or "css" will be added
      as separate attributes"""

    def get_placeholder(self, content_type_id, object_id, slot):
        try:
            placeholder = Placeholder.objects.get(
                content_type_id=content_type_id, object_id=object_id, slot=slot
            )
        except Placeholder.DoesNotExist:
            raise Http404
        return placeholder

    def get(self, request, language, content_type_id, object_id, slot, format=None):
        placeholder = self.get_placeholder(content_type_id, object_id, slot)
        if placeholder is None:
            raise Http404
        source = (
            placeholder.content_type.model_class()
            .objects.filter(pk=placeholder.object_id)
            .first()
        )
        if source is None:
            raise Http404
        # Check if the user has permission to view the page (should the placeholder be on a page)
        if isinstance(source, PageContent) and not user_can_view_page(
            request.user, source.page
        ):
            raise Http404
        serializer = PlaceholderSerializer(
            request, placeholder, language, read_only=True
        )
        return Response(serializer.data)



from django.urls import path
from rest_framework.urlpatterns import format_suffix_patterns
from . import views

urlpatterns = [
    path("", views.LanguageList.as_view(), name="cms-language-list"),
    path("<slug:language>/pages", views.PageList.as_view(), name="cms-page-list"),
    path("<slug:language>/pages/", views.PageDetail.as_view(), name="cms-page-root"),
    path("<slug:language>/pages/<path:path>/", views.PageDetail.as_view(), name="cms-page-detail"),
    path(
        "<slug:language>/placeholders/<int:content_type_id>/<int:object_id>/<str:slot>/",
        views.PlaceholderDetail.as_view(),
        name="cms-placeholder-detail",
    ),
]

urlpatterns = format_suffix_patterns(urlpatterns)
